"""This file contains all the Qualys API URLs used by qualyspy."""

about = "/msp/about.php/"
host_list_vm_detection = "/api/3.0/fo/asset/host/vm/detection/"
host_list = "/api/4.0/fo/asset/host/"
knowledgebase = "/api/3.0/fo/knowledge_base/vuln/"
ignore_vuln = "/api/2.0/fo/ignore_vuln/index.php/"
launch_vm_scan = "/api/2.0/fo/scan/"
map_report_list = "/msp/map_report_list.php"
download_saved_map_report = "/msp/map_report.php"

gateway_auth = "/auth/"

asset_details = "/rest/2.0/get/am/asset"
all_asset_details = "/rest/2.0/search/am/asset"

create_tag = "/qps/rest/2.0/create/am/tag"
update_tag = "/qps/rest/2.0/update/am/tag"
search_tags = "/qps/rest/2.0/search/am/tag"
count_tags = "/qps/rest/2.0/count/am/tag"
delete_tag = "/qps/rest/2.0/delete/am/tag"
get_asset_info = "/qps/rest/2.0/get/am/asset"
update_asset = "/qps/rest/2.0/update/am/asset"
search_assets = "/qps/rest/2.0/search/am/asset"

add_bulk_external_sites = "/certview/v1/externalSites/bulkAdd"
